/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004-2005  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: tstore.c,v 1.47 2008/04/13 09:43:40 kasperd Exp $
*/

#include "tstore.h"
#include "test.h"

struct store *store;

int zblock(uint32_t *buf)
{
  int i;
  for (i=0;i<128;++i)
    if (buf[i]) return 0;
  return 1;
}

FILE *hash_file;
uint64_t filesize=0;

wtree_t allocated_node=NULL;

wtree_t get_allocated_node()
{
  if (!allocated_node) allocated_node=allocnode();
  return allocated_node;
}

wtree_t put_allocated_node()
{
  wtree_t res=insert(&hsec_tree,allocated_node);
  if (!res) {
    res=allocated_node;
    allocated_node=NULL;
  }
  return res;
}

#define MAX_BATCH 19531250
struct si{
  uint64_t *l;
  union {
    uint8_t *d;
    uint64_t l;
  } u;
} sec_info[MAX_BATCH+2];

/* FIXME: The batch size code needs to be improved */
#if defined(_SC_PHYS_PAGES) && defined(_SC_PAGE_SIZE)
int compute_batch_size()
{
  int x=1,y=sysconf(_SC_PHYS_PAGES)*(sysconf(_SC_PAGE_SIZE)/1024);
  char *BATCHSIZE=getenv("BATCHSIZE");
  if (BATCHSIZE) {
    int i=atoi(BATCHSIZE);
    if (i>42) y=i;
  }
  if (y>MAX_BATCH) return MAX_BATCH;
  if (y<42) {
    fprintf(stderr,"You need at least 64MB of RAM\n");
    exit(EXIT_FAILURE);
  }
  while(y>12) {
    y/=5;
    x*=5;
  }
  return x*y;
}
#else
#define compute_batch_size() 78125
#endif

int load_sector(int fd, struct si *si)
{
  wtree_t node=get_allocated_node();
  int r,z;
  r=read_and_hash(fd,datap(node));
  if((r<1)||(r>512)) return r;
  z=zblock((void*)(datap(node)+16));
  node=put_allocated_node();
  if(z) {
    si->l=&(si->u.l);
    si->u.l=0;
  } else {
    si->l=locationp(node);
    si->u.d=datap(node);
  }
  return r;
}

struct block_num_queue {
  struct block_num_queue *next,*prev;
  int count;
  uint64_t numbers[FANOUT];
} *queue_head;

void put_block_number(struct block_num_queue *head,
		      uint64_t blocknumber)
{
  if ((head->prev == head)||(head->prev->count==FANOUT)) {
    struct block_num_queue *new;
    NEW(new);
    new->prev=head->prev;
    new->next=head;
    new->count=0;
    head->prev->next=new;
    head->prev=new;
  }
  head->prev->numbers[head->prev->count++]=blocknumber;
}

void init_queue()
{
  NEW(queue_head);
  queue_head->next=queue_head->prev=queue_head;
  queue_head->count=-1;
}

void batch2secnum()
{
  uint64_t i;

  assert(fseek(hash_file,16,SEEK_SET)==0);

  for (i=1;i<store->store_size;++i) {
    uint8_t hash_sector_buf[HBS];
    wtree_t t;
    wtree_key_t hash_value;
    
    if (fread(&hash_value,1,sizeof(wtree_key_t),hash_file)!=sizeof(wtree_key_t)) errstop(store->h32name);

    t=find_hash(hsec_tree,hash_value);
    if (t) {
      /* FIXME: Use pread instead */
      lseek(store->fd,512*i,SEEK_SET);
      switch(read_and_hash(store->fd,hash_sector_buf)) {
      case -1:
	errstop("read");
      case 512:
	break;
      case 0:
	fprintf(stderr,"Short read\n");
	exit(EXIT_FAILURE);
      }
      t=find(t,hash_sector_buf);
      if (t) {
	uint64_t *lp=locationp(t);
	assert(!*lp);
	*lp=i;
      }
    }
  }
}

void secnum2queue(struct block_num_queue *q, uint64_t batchsize)
{
  uint64_t i;
  assert(fseek(hash_file,(store->store_size+3)*sizeof(wtree_key_t),SEEK_SET)==0);
  lseek(store->fd,store->store_size*512,SEEK_SET);
  for (i=0;i<batchsize;++i) {
    uint64_t *l=sec_info[i].l;
    if (l != &(sec_info[i].u.l)) {
      if (!*l) {
	uint8_t *dp=sec_info[i].u.d;
	assert(write(store->fd,dp+16,512)==512);
	assert(fwrite(dp,1,sizeof(wtree_key_t),hash_file)==sizeof(wtree_key_t));
	*l=store->store_size++;
      }
    }
    put_block_number(q,*l);
  }
}

struct timespec fd_mtime(int fd)
{
  struct stat st_buf;
  struct timespec r;
  assert(fstat(fd,&st_buf)==0);
#if defined(_STAT_VER_LINUX) && defined(st_mtime)
  r=st_buf.st_mtim;
#else
  r.tv_sec=st_buf.st_mtime;
  r.tv_nsec=0;
#endif
  return r;
}

struct timespec ts_tod()
{
  struct timeval tv;
  struct timespec ts;
  if (gettimeofday(&tv,NULL)) errstop("gettimeofday");
  ts.tv_sec=tv.tv_sec;
  ts.tv_nsec=tv.tv_usec*1000;
  return ts;
}

int convert(char *s, struct timespec *ts)
{
  char *r;
  int v,l;
  if (!*s) return -1;
  ts->tv_sec=strtoll(s,&r,10);
  ts->tv_nsec=0;
  if (!*r) return 0;
  if ((*r != '.')&&(*r != ',')) return -1;
  ++r;
  l=strlen(r);
  if (l>9) return -1;
  v=strtoll(r,&r,10);
  if (*r) return -1;
  while (l<9) {
    v*=10;
    l++;
  }
  ts->tv_nsec=v;
  return 0;
}

void progress(int c)
{
  putchar(c);
  fflush(stdout);
}

int main(int argc, char ** argv)
{
  struct timespec ts;
  int DEFAULT_BATCH=compute_batch_size();

  char formbuf[42];

  char *progbuf=getenv("TSTORE");
  if (progbuf) sprintf(formbuf,"%%%dlld",strlen(progbuf));

  init_queue();

  if ((argc != 3)&&(argc != 4)) {
    fprintf(stderr,"Uage: %s <store> <file> [timestamp]\n",argv[0]);
    return 1;
  }

  if ((argc==4)&&strcmp(argv[3],"-")) {
    if (convert(argv[3],&ts)) {
      fprintf(stderr,"%s is not a valid timestamp\n",argv[3]);
      return EXIT_FAILURE;
    }
  }

  if (strlen(argv[2])>TS_MAX_NAME_LEN) {
    fprintf(stderr,"%s: Name too long\n",argv[2]);
    return 1;
  }

  store=open_store(argv[1],O_RDWR,CHECK_LOCK);
  if (!store) errstop(argv[1]);

  hash_file=fopen(store->h32name,"r+");
  if (!hash_file) errstop(store->h32name);

  if (fread(hash_key,1,16,hash_file)!=16) errstop(store->h32name);

  while(1) {

    int r;
    int batch_size=0;

    progress('[');
    assert(hsec_tree == NULL);

    while((batch_size < DEFAULT_BATCH)&&(r=load_sector(0,sec_info+batch_size))) {
      if ((r>512)||(r<0)) errstop("read");
      if (lock_store(store)) errstop(store->name);
      filesize += r;
      if (progbuf) sprintf(progbuf,formbuf,filesize);
      assert(batch_size++ < MAX_BATCH);
    }

    if (!batch_size) break;

    progress('.');
    batch2secnum();
    progress('.');
    secnum2queue(queue_head,batch_size);
    progress('.');

    allocated_node = NULL;
    discard_tree();

    progress(']');
  }

  if (filesize == 0) return 0;

  while (queue_head->next->count != 1) {
    struct block_num_queue *q=queue_head;
    int batch_size=0;
    progress('=');
    init_queue();

    for (batch_size=0; q->next != q; ++batch_size) {
      /* Each queue node is written as one indirection block */
      int count=q->next->count;
      uint64_t *numbers=q->next->numbers;
      int z=1,s=0,i;
      assert(count>0);
      /* First verify if this block is all zero */
      for (i=0;i<count;++i) {
	if (numbers[i]) {
	  z=0;
	  break;
	}
      }
      /* Next verify if this could be a sequential block */
      if (z||(numbers[0]>=0xC000000000ULL)) {
	s=0;
      } else {
	int delta=0;
	switch(numbers[0]>>38) {
	case 0: delta=1; break;
	case 1: delta=FANOUT; break;
	case 2: delta=FANOUT*FANOUT; break;
	}
	assert(delta);
	s=1;
	for(i=1;i<count;++i) {
	  if (numbers[i]!=numbers[i-1]+delta) {
	    s=0;
	    break;
	  }
	}
      }
      if (z||s) {
	/* Use shortcut for zero blocks and sequential blocks */
	sec_info[batch_size].l=&(sec_info[batch_size].u.l);
	sec_info[batch_size].u.l=z?0ULL:numbers[0]+0x4000000000ULL;
      } else {
	/* No shortcut possible, I must actually write
	 * this block if it is not already there.
	 */
	wtree_t node=get_allocated_node();
	uint8_t *buf=datap(node);
	for (i=0;i<count;++i) PUT_BNUM(buf+(16+5*i),numbers[i]);
	hash_buf(buf,count*5);
	node=put_allocated_node();
	sec_info[batch_size].u.d=datap(node);
	sec_info[batch_size].l=locationp(node);
      }
      q->next=q->next->next;
    }

    assert(batch_size);

    progress('=');
    batch2secnum();
    progress('=');
    secnum2queue(queue_head,batch_size);
    progress('=');

    allocated_node = NULL;
    discard_tree();

    progress('|');
  }

  /* Write the file header record */
  {
    wtree_t node=get_allocated_node();
    uint8_t *buf=datap(node)+16;
  
    memset(buf,0,512);
    assert(pread(store->fd,buf+STORESIZE_OFFSET,10,STORESIZE_OFFSET)==10);
    PUT_BNUM(buf+TREEPTR_OFFSET,queue_head->next->numbers[0]);
    PUT_SIZE(buf+FILESIZE_OFFSET,filesize);
    strcpy(buf+FILENAME_OFFSET,argv[2]);
    ts=(argc==4)?(strcmp(argv[3],"-")?ts:fd_mtime(0)):ts_tod();
    PUT_TIME(buf+SECONDS_OFFSET,ts.tv_sec);
    PUT_NTIM(buf+NANOSEC_OFFSET,ts.tv_nsec);

    progress('.');
    hash_buf(datap(node),512);
    node=put_allocated_node();
    sec_info[0].u.d=datap(node);
    sec_info[0].l=locationp(node);

    batch2secnum();
    progress('.');
    secnum2queue(queue_head,1);
    progress(']');

    assert(queue_head->next->count==2);
    
    PUT_BNUM(buf,store->store_size);
    PUT_BNUM(buf+5,queue_head->next->numbers[1]);
    assert(pwrite(store->fd,buf,10,STORESIZE_OFFSET)==10);
    progress('\n');

    close_store(store);

    return 0;
  }
}
