/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004-2005  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: wtree.c,v 1.18 2008/04/13 09:43:40 kasperd Exp $
*/

#include "tstore.h"
#include "test.h"

union data {
  wtree_key_t k;
  uint8_t d[HBS];
};

/* A weight balanced tree with a few extra data fields */
struct wtree {
  wtree_t left,right;
#ifdef WEIGHT
  int w;
#endif
  uint64_t location;
  wtree_key_t key;
  union data *data;
};

wtree_t hsec_tree=NULL;

static inline wtree_t R(wtree_t t)
{
  return t?t->right:NULL;
}
static inline wtree_t L(wtree_t t)
{
  return t?t->left:NULL;
}
#ifdef WEIGHT
static inline int W(wtree_t t)
{
  return t?t->w:0;
}

static inline void split(wtree_t *a,wtree_t t)
{
  if (W(L(t))>W(R(t))) {
    a[0]=t->left->left;
    a[1]=t->left;
    a[2]=t->left->right;
    a[3]=t;
    a[4]=t->right;
  } else {
    a[0]=t->left;
    a[1]=t;
    a[2]=t->right->left;
    a[3]=t->right;
    a[4]=t->right->right;
  }
}

void rebalance(wtree_t *tp)
{
  wtree_t flat[7];
  int rebuild;
  do {
    int WL=W(L(*tp)),WR=W(R(*tp));
    assert(W(*tp)==WL+WR+1);
    rebuild=0;
    if (WR>5*(WL+1)) {
      flat[0]=(*tp)->left;
      flat[1]=(*tp);
      split(flat+2,(*tp)->right);
      rebuild=1;
    }
    if (WL>5*(WR+1)) {
      split(flat,(*tp)->left);
      flat[5]=(*tp);
      flat[6]=(*tp)->right;
      rebuild=1;
    }
    if (rebuild) {
      flat[1]->left=flat[0];
      flat[1]->right=flat[2];
      flat[1]->w=W(flat[0])+W(flat[2])+1;
      
      flat[5]->left=flat[4];
      flat[5]->right=flat[6];
      flat[5]->w=W(flat[4])+W(flat[6])+1;
      
      flat[3]->left=flat[1];
      flat[3]->right=flat[5];
      flat[3]->w=W(flat[1])+W(flat[5])+1;
      
      *tp=flat[3];
    }
  } while(rebuild);
}
#endif

static inline int nodestringcmp(wtree_t t, uint8_t *s)
{
  wtree_key_t h;
  memcpy(&h,s,sizeof(wtree_key_t));
  if (t->key<h) return -1;
  if (t->key>h) return 1;
  return memcmp(t->data,s,HBS);
}

wtree_t insert(wtree_t *tp, wtree_t newnode)
{
  if (*tp == NULL) {
    (*tp)=newnode;
    newnode->key=newnode->data->k;
    newnode->left=newnode->right=NULL;
#ifdef WEIGHT
    newnode->w=1;
#endif
    return NULL;
  } else {
    int c=nodestringcmp(*tp,newnode->data->d);
    wtree_t r=*tp;
#ifdef WEIGHT
    assert(W(*tp)==W(L(*tp))+W(R(*tp))+1);
#endif
    if (c<0) r=insert(&((*tp)->right),newnode);
    if (c>0) r=insert(&((*tp)->left),newnode);
#ifdef WEIGHT
    if (((*tp)->w+=(r==NULL))>8) rebalance(tp);
#endif
    return r;
  }
}

wtree_t find(wtree_t t, uint8_t *elm)
{
  int c;
  if (t==NULL) return t;
  c=nodestringcmp(t,elm);
  if (c<0) return find(t->right,elm);
  if (c>0) return find(t->left,elm);
  return t;
}

wtree_t find_hash(wtree_t t, wtree_key_t h)
{
  if (t==NULL) return t;
  if (t->key<h) return find_hash(t->right,h);
  if (t->key>h) return find_hash(t->left,h);
  return t;
}

void **slabchain=NULL;
int avail_nodes=0;
int avail_buffers=0;

void discard_tree()
{
  avail_nodes=avail_buffers=0;
  hsec_tree = NULL;
  while (slabchain) {
    void **t=slabchain;
    slabchain=*t;
    munmap(t,16384);
  }
}

void *get_slab()
{
  void **p=mmap(NULL,16384,
		PROT_READ|PROT_WRITE,
		MAP_PRIVATE|MAP_ANONYMOUS,
		0,0);
  if (p==MAP_FAILED) errstop("mmap");
  *p=slabchain;
  slabchain=p;
  return 16+(char*)p;
}

uint64_t *locationp(wtree_t t)
{
  return &(t->location);
}

uint8_t *datap(wtree_t t)
{
  return t->data->d;
}

wtree_t allocnode()
{
  static union data *bp=NULL;
  static wtree_t r=NULL;
  if (!avail_nodes) {
    r=get_slab();
    avail_nodes=(16*1023)/sizeof(*r);
  }
  --avail_nodes;

  if (!avail_buffers) {
    bp=get_slab();
    avail_buffers=31; /* 16 + 31*(512+16) = 16KB */
  }
  --avail_buffers;
  r->data=bp++;

  r->location=0;
#ifdef WEIGHT
  r->w=-42;
#endif
  r->left=r->right=NULL;
  return r++;
}
